/* $Id: cfb.c,v 1.2 2002/01/25 08:14:54 levitte Exp $ */

/* ====================================================================
 * Copyright (c) 1999 Levitte Programming.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *      This product includes software developed by Levitte Programming
 *      and its contributors.
 *
 * 4. The names "RLCRYPTO" and "Levitte Programming" or the names of its
 *    contributors must not be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *    For written permission, please contact levitte@lp.se.
 *
 * THIS SOFTWARE IS PROVIDED BY LEVITTE PROGRAMMING AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL LEVITTE
 * PROGRAMMING OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 *
 */

#include <stdlib.h>
#include <string.h>
#include "../algo_int.h"

struct algorithm_context_st
{
  unsigned char iv[64];
  size_t iv_len;
};

static size_t cfb_ctx_size(void)
{
  return sizeof(ALGO_CTX);
}

static ALGO_CTX *cfb_alloc(void)
{
  return (ALGO_CTX *)malloc(cfb_ctx_size());
}

static void cfb_free(ALGO_CTX *ctx)
{
  free(ctx);
}

static int cfb_init(ALGO_CTX *ctx, const unsigned char *ivec, size_t iveclen)
{
  memset(ctx->iv, 0, sizeof(ctx->iv));
  memcpy(ctx->iv, ivec, iveclen);
  if (iveclen > 64)
    return ALGO_E_TOO_LARGE_BLOCK;
  ctx->iv_len = iveclen;
  return ALGO_E_OK;
}

static void cfb_clear(ALGO_CTX *ctx)
{
  memset(ctx, 0, cfb_ctx_size());
}

static int cfb1_enc_dec(ALGO_CTX *ctx,
			const unsigned char *input, unsigned char *output, size_t length,
			int encrypt,
			int (*sym_enc_dec)(ALGO_CTX *ctx,
					   const unsigned char *input,
					   unsigned char *output, size_t length,
					   int encrypt),
			ALGO_CTX *sym_data)
{
  int status = ALGO_E_OK;
  size_t n;
  unsigned char tmp[64];
  register size_t iv_len = ctx->iv_len;
  register unsigned char current_omask = 0x7f;
  register unsigned char current_ishift = 0;

  while (length)
    {
      status = sym_enc_dec(sym_data, ctx->iv, tmp, iv_len,
			   OPENSSL_ENCRYPT);
      if (status != ALGO_E_OK)
	return status;
      *output = (*output & current_omask)
	| ((*input & ~current_omask) ^ ((tmp[0] & 0x80) >> current_ishift));
      for(n=0; n < iv_len-1; n++)
	ctx->iv[n] = (ctx->iv[n] << 1)
	  | (ctx->iv[n+1] >> 7);
      if (encrypt)
	ctx->iv[n] = (ctx->iv[n] << 1)
	  | ((*output >> (7 - current_ishift)) & 0x01);
      else
	ctx->iv[n] = (ctx->iv[n] << 1)
	  | ((*input >> (7 - current_ishift)) & 0x01);
      if (current_ishift == 7)
	{
	  output++;
	  input++;
	  length--;
	  current_omask = 0x7f;
	  current_ishift = 0;
	}
      else
	{
	  current_omask = (current_omask >> 1) | 0x80;
	  current_ishift++;
	}
    }
  return status;
}

static int cfb8_enc_dec(ALGO_CTX *ctx,
			const unsigned char *input, unsigned char *output, size_t length,
			int encrypt,
			int (*sym_enc_dec)(ALGO_CTX *ctx,
					   const unsigned char *input,
					   unsigned char *output, size_t length,
					   int encrypt),
			ALGO_CTX *sym_data)
{
  int status = ALGO_E_OK;
  unsigned char tmp[64];
  register size_t iv_len = ctx->iv_len;

  while (length)
    {
      status = sym_enc_dec(sym_data, ctx->iv, tmp, iv_len,
			   OPENSSL_ENCRYPT);
      if (status != ALGO_E_OK)
	return status;
      *output = *input ^ tmp[0];
      memcpy(ctx->iv, &ctx->iv[1], iv_len - 1);
      if (encrypt)
	ctx->iv[iv_len - 1] = *output;
      else
	ctx->iv[iv_len - 1] = *input;
      output++;
      input++;
      length--;
    }
  return status;
}

static int cfb64_enc_dec(ALGO_CTX *ctx,
			 const unsigned char *input, unsigned char *output, size_t length,
			 int encrypt,
			 int (*sym_enc_dec)(ALGO_CTX *ctx,
					    const unsigned char *input,
					    unsigned char *output, size_t length,
					    int encrypt),
			 ALGO_CTX *sym_data)
{
  int status = ALGO_E_OK;
  size_t n;
  unsigned char tmp[64];
  register size_t iv_len = ctx->iv_len;

  if (iv_len < 8 || iv_len % 8 != 0)
    return ALGO_E_INVALID_BLOCK_SIZE;

  if (length % iv_len != 0)
    return ALGO_E_NOT_EXACT_BLOCKS;

  while (length)
    {
      status = sym_enc_dec(sym_data, ctx->iv, tmp, iv_len,
			   OPENSSL_ENCRYPT);
      if (status != ALGO_E_OK)
	return status;
      for(n = 0; n < 8; n++)
	output[n] = input[n] ^ tmp[n];
      memcpy(ctx->iv, &ctx->iv[8], iv_len - 8);
      if (encrypt)
	memcpy(&ctx->iv[iv_len - 8], output, 8);
      else
	memcpy(&ctx->iv[iv_len - 8], input, 8);
      output += 8;
      input += 8;
      length -= 8;
    }
  return status;
}

static int cfb128_enc_dec(ALGO_CTX *ctx,
			  const unsigned char *input, unsigned char *output, size_t length,
			  int encrypt,
			  int (*sym_enc_dec)(ALGO_CTX *ctx,
					     const unsigned char *input,
					     unsigned char *output, size_t length,
					     int encrypt),
			  ALGO_CTX *sym_data)
{
  int status = ALGO_E_OK;
  size_t n;
  unsigned char tmp[64];
  register size_t iv_len = ctx->iv_len;

  if (iv_len < 16 || iv_len % 16 != 0)
    return ALGO_E_INVALID_BLOCK_SIZE;

  if (length % iv_len != 0)
    return ALGO_E_NOT_EXACT_BLOCKS;

  while (length)
    {
      status = sym_enc_dec(sym_data, ctx->iv, tmp, iv_len,
			   OPENSSL_ENCRYPT);
      if (status != ALGO_E_OK)
	return status;
      for(n = 0; n < 16; n++)
	output[n] = input[n] ^ tmp[n];
      memcpy(ctx->iv, &ctx->iv[16], iv_len - 16);
      if (encrypt)
	memcpy(&ctx->iv[iv_len - 16], output, 16);
      else
	memcpy(&ctx->iv[iv_len - 16], input, 16);
      output += 16;
      input += 16;
      length -= 16;
    }
  return status;
}

static int cfb_cpy_ivec(ALGO_CTX *ctx, unsigned char *ivec, size_t ivecsize,
			size_t *iveclen)
{
  if (ivecsize < ctx->iv_len)
    return ALGO_E_INVALID_IVEC_SIZE;
  memcpy(ivec, ctx->iv, ctx->iv_len);
  *iveclen = ctx->iv_len;
  return ALGO_E_OK;
}

static const MODE_METHOD cfb1_method =
  {
    "CFB1",
    0,
    1,
    &cfb_ctx_size,
    &cfb_alloc,
    &cfb_free,
    &cfb_init,
    &cfb_clear,
    &cfb1_enc_dec,
    &cfb_cpy_ivec
  };

const MODE_METHOD *CFB1(void)
{
  return &cfb1_method;
}

static const MODE_METHOD cfb8_method =
  {
    "CFB8",
    0,
    1,
    &cfb_ctx_size,
    &cfb_alloc,
    &cfb_free,
    &cfb_init,
    &cfb_clear,
    &cfb8_enc_dec,
    &cfb_cpy_ivec
  };

const MODE_METHOD *CFB8(void)
{
  return &cfb8_method;
}

static const MODE_METHOD cfb64_method =
  {
    "CFB64",
    0,
    1,
    &cfb_ctx_size,
    &cfb_alloc,
    &cfb_free,
    &cfb_init,
    &cfb_clear,
    &cfb64_enc_dec,
    &cfb_cpy_ivec
  };

const MODE_METHOD *CFB64(void)
{
  return &cfb64_method;
}

static const MODE_METHOD cfb128_method =
  {
    "CFB128",
    0,
    1,
    &cfb_ctx_size,
    &cfb_alloc,
    &cfb_free,
    &cfb_init,
    &cfb_clear,
    &cfb128_enc_dec,
    &cfb_cpy_ivec
  };

const MODE_METHOD *CFB128(void)
{
  return &cfb128_method;
}
